@@ -22,12 +22,12 @@ import java.nio.charset.StandardCharsets;
 import java.util.EnumSet;
 import java.util.concurrent.atomic.AtomicBoolean;
 
-import javax.servlet.ReadListener;
-import javax.servlet.RequestDispatcher;
-import javax.servlet.ServletException;
-import javax.servlet.SessionTrackingMode;
-import javax.servlet.WriteListener;
-import javax.servlet.http.HttpServletResponse;
+import jakarta.servlet.ReadListener;
+import jakarta.servlet.RequestDispatcher;
+import jakarta.servlet.ServletException;
+import jakarta.servlet.SessionTrackingMode;
+import jakarta.servlet.WriteListener;
+import jakarta.servlet.http.HttpServletResponse;
 
 import org.apache.catalina.Authenticator;
 import org.apache.catalina.Context;
@@ -46,6 +46,7 @@ import org.apache.tomcat.util.ExceptionUtils;
 import org.apache.tomcat.util.buf.B2CConverter;
 import org.apache.tomcat.util.buf.ByteChunk;
 import org.apache.tomcat.util.buf.CharChunk;
+import org.apache.tomcat.util.buf.HexUtils;
 import org.apache.tomcat.util.buf.MessageBytes;
 import org.apache.tomcat.util.http.ServerCookie;
 import org.apache.tomcat.util.http.ServerCookies;
@@ -66,16 +67,13 @@ public class CoyoteAdapter implements Adapter {
 
     // -------------------------------------------------------------- Constants
 
-    private static final String POWERED_BY = "Servlet/4.0 JSP/2.3 " + "(" + ServerInfo.getServerInfo() + " Java/" +
+    private static final String POWERED_BY = "Servlet/6.0 JSP/3.1 " + "(" + ServerInfo.getServerInfo() + " Java/" +
             System.getProperty("java.vm.vendor") + "/" + System.getProperty("java.runtime.version") + ")";
 
     private static final EnumSet<SessionTrackingMode> SSL_ONLY = EnumSet.of(SessionTrackingMode.SSL);
 
     public static final int ADAPTER_NOTES = 1;
 
-    protected static final boolean ALLOW_BACKSLASH = Boolean
-            .parseBoolean(System.getProperty("org.apache.catalina.connector.CoyoteAdapter.ALLOW_BACKSLASH", "false"));
-
 
     // ----------------------------------------------------------- Constructors
 
@@ -139,7 +137,7 @@ public class CoyoteAdapter implements Adapter {
                 }
             } else if (status == SocketEvent.ERROR) {
                 // An I/O error occurred on a non-container thread which means
-                // that the socket needs to be closed so set success as false to
+                // that the socket needs to be closed so set success to false to
                 // trigger a close
                 success = false;
                 Throwable t = (Throwable) req.getAttribute(RequestDispatcher.ERROR_EXCEPTION);
@@ -179,8 +177,6 @@ public class CoyoteAdapter implements Adapter {
                         }
                     } catch (Throwable t) {
                         ExceptionUtils.handleThrowable(t);
-                        // Allow the error handling to write to the response
-                        response.setSuspended(false);
                         // Need to trigger the call to AbstractProcessor.setErrorState()
                         // before the listener is called so the listener can call complete
                         // Therefore no need to set success=false as that would trigger a
@@ -214,8 +210,6 @@ public class CoyoteAdapter implements Adapter {
                         }
                     } catch (Throwable t) {
                         ExceptionUtils.handleThrowable(t);
-                        // Allow the error handling to write to the response
-                        response.setSuspended(false);
                         // Need to trigger the call to AbstractProcessor.setErrorState()
                         // before the listener is called so the listener can call complete
                         // Therefore no need to set success=false as that would trigger a
@@ -239,8 +233,8 @@ public class CoyoteAdapter implements Adapter {
 
             if (request.isAsyncDispatching()) {
                 connector.getService().getContainer().getPipeline().getFirst().invoke(request, response);
-                if (response.isError()) {
-                    Throwable t = (Throwable) request.getAttribute(RequestDispatcher.ERROR_EXCEPTION);
+                Throwable t = (Throwable) request.getAttribute(RequestDispatcher.ERROR_EXCEPTION);
+                if (t != null) {
                     asyncConImpl.setErrorState(t, true);
                 }
             }
@@ -255,16 +249,17 @@ public class CoyoteAdapter implements Adapter {
             AtomicBoolean error = new AtomicBoolean(false);
             res.action(ActionCode.IS_ERROR, error);
             if (error.get()) {
-                if (request.isAsyncCompleting() || request.isAsyncDispatching()) {
-                    // Connection will be forcibly closed which will prevent completion/dispatch happening at the usual
-                    // point. Trigger post-processing here.
+                if (request.isAsyncCompleting()) {
+                    // Connection will be forcibly closed which will prevent
+                    // completion happening at the usual point. Need to trigger
+                    // call to onComplete() here.
                     res.action(ActionCode.ASYNC_POST_PROCESS, null);
                 }
                 success = false;
             }
-        } catch (IOException ioe) {
-            // Issues that should be logged will have already been logged
+        } catch (IOException e) {
             success = false;
+            // Ignore
         } catch (Throwable t) {
             ExceptionUtils.handleThrowable(t);
             success = false;
@@ -277,8 +272,8 @@ public class CoyoteAdapter implements Adapter {
             // Access logging
             if (!success || !request.isAsync()) {
                 long time = 0;
-                if (req.getStartTime() != -1) {
-                    time = System.currentTimeMillis() - req.getStartTime();
+                if (req.getStartTimeNanos() != -1) {
+                    time = System.nanoTime() - req.getStartTimeNanos();
                 }
                 Context context = request.getContext();
                 if (context != null) {
@@ -375,8 +370,8 @@ public class CoyoteAdapter implements Adapter {
                 response.finishResponse();
             }
 
-        } catch (IOException ignore) {
-            // Issues that should be logged will have already been logged
+        } catch (IOException e) {
+            // Ignore
         } finally {
             AtomicBoolean error = new AtomicBoolean(false);
             res.action(ActionCode.IS_ERROR, error);
@@ -403,7 +398,7 @@ public class CoyoteAdapter implements Adapter {
                 // The other possibility is that an error occurred early in
                 // processing and the request could not be mapped to a Context.
                 // Log via the host or engine in that case.
-                long time = System.currentTimeMillis() - req.getStartTime();
+                long time = System.nanoTime() - req.getStartTimeNanos();
                 if (context != null) {
                     context.logAccess(request, response, time, false);
                 } else if (response.isError()) {
@@ -597,7 +592,7 @@ public class CoyoteAdapter implements Adapter {
 
         // Check for ping OPTIONS * request
         if (undecodedURI.equals("*")) {
-            if (req.method().equals("OPTIONS")) {
+            if (req.method().equalsIgnoreCase("OPTIONS")) {
                 StringBuilder allow = new StringBuilder();
                 allow.append("GET, HEAD, POST, PUT, DELETE, OPTIONS");
                 // Trace if allowed
@@ -616,11 +611,17 @@ public class CoyoteAdapter implements Adapter {
         MessageBytes decodedURI = req.decodedURI();
 
         // Filter CONNECT method
-        if (req.method().equals("CONNECT")) {
+        if (req.method().equalsIgnoreCase("CONNECT")) {
             response.sendError(HttpServletResponse.SC_NOT_IMPLEMENTED, sm.getString("coyoteAdapter.connect"));
         } else {
             // No URI for CONNECT requests
             if (undecodedURI.getType() == MessageBytes.T_BYTES) {
+                if (connector.getRejectSuspiciousURIs()) {
+                    if (checkSuspiciousURIs(undecodedURI.getByteChunk())) {
+                        response.sendError(400, sm.getString("coyoteAdapter.invalidURI"));
+                    }
+                }
+
                 // Copy the raw URI to the decodedURI
                 decodedURI.duplicate(undecodedURI);
 
@@ -631,39 +632,35 @@ public class CoyoteAdapter implements Adapter {
                 // %xx decoding of the URL
                 try {
                     req.getURLDecoder().convert(decodedURI.getByteChunk(),
-                            connector.getEncodedSolidusHandlingInternal(),
-                            connector.getEncodedReverseSolidusHandlingInternal());
+                            connector.getEncodedSolidusHandlingInternal());
                 } catch (IOException ioe) {
                     response.sendError(400, sm.getString("coyoteAdapter.invalidURIWithMessage", ioe.getMessage()));
                 }
                 // Normalization
-                if (normalize(req.decodedURI())) {
+                if (normalize(req.decodedURI(), connector.getAllowBackslash())) {
                     // Character decoding
                     convertURI(decodedURI, request);
-                    // Check that the URI is still normalized
-                    // Note: checkNormalize is deprecated because the test is no
-                    // longer required in Tomcat 10 onwards and has been
-                    // removed
-                    if (!checkNormalize(req.decodedURI())) {
-                        response.sendError(400, "Invalid URI");
-                    }
+                    // URIEncoding values are limited to US-ASCII supersets.
+                    // Therefore it is not necessary to check that the URI remains
+                    // normalized after character decoding
                 } else {
                     response.sendError(400, sm.getString("coyoteAdapter.invalidURI"));
                 }
             } else {
                 /*
                  * The URI is chars or String, and has been sent using an in-memory protocol handler. The following
-                 * assumptions are made:
-                 *
-                 * - req.requestURI() has been set to the 'original' non-decoded, non-normalized URI that includes path
-                 * parameters (if any)
-                 *
-                 * - req.decodedURI() has been set to the decoded, normalized form of req.requestURI() with any path
-                 * parameters removed
-                 *
-                 * - 'suspicious' URI filtering, if required, has already been performed
+                 * assumptions are made: - req.requestURI() has been set to the 'original' non-decoded, non-normalized
+                 * URI - req.decodedURI() has been set to the decoded, normalized form of req.requestURI() -
+                 * 'suspicious' URI filtering - if required - has already been performed
                  */
                 decodedURI.toChars();
+                // Remove all path parameters; any needed path parameter should be set
+                // using the request object rather than passing it in the URL
+                CharChunk uriCC = decodedURI.getCharChunk();
+                int semicolon = uriCC.indexOf(';');
+                if (semicolon > 0) {
+                    decodedURI.setChars(uriCC.getBuffer(), uriCC.getStart(), semicolon);
+                }
             }
         }
 
@@ -693,7 +690,7 @@ public class CoyoteAdapter implements Adapter {
         }
 
         while (mapRequired) {
-            // This will map the latest version by default
+            // This will map the the latest version by default
             connector.getService().getMapper().map(serverName, decodedURI, version, request.getMappingData());
 
             // If there is no context at this point, either this is a 404
@@ -782,8 +779,8 @@ public class CoyoteAdapter implements Adapter {
                 // point.
                 try {
                     Thread.sleep(1000);
-                } catch (InterruptedException ignore) {
-                    // Should never happen but, if it does, just continue looping
+                } catch (InterruptedException e) {
+                    // Should never happen
                 }
                 // Reset mapping
                 request.getMappingData().recycle();
@@ -813,9 +810,9 @@ public class CoyoteAdapter implements Adapter {
         }
 
         // Filter TRACE method
-        if (!connector.getAllowTrace() && req.method().equals("TRACE")) {
+        if (!connector.getAllowTrace() && req.method().equalsIgnoreCase("TRACE")) {
             Wrapper wrapper = request.getWrapper();
-            StringBuilder header = null;
+            String header = null;
             if (wrapper != null) {
                 String[] methods = wrapper.getServletMethods();
                 if (methods != null) {
@@ -824,15 +821,15 @@ public class CoyoteAdapter implements Adapter {
                             continue;
                         }
                         if (header == null) {
-                            header = new StringBuilder(method);
+                            header = method;
                         } else {
-                            header.append(", ").append(method);
+                            header += ", " + method;
                         }
                     }
                 }
             }
             if (header != null) {
-                res.addHeader("Allow", header.toString());
+                res.addHeader("Allow", header);
             }
             response.sendError(405, sm.getString("coyoteAdapter.trace"));
             // Safe to skip the remainder of this method.
@@ -868,7 +865,7 @@ public class CoyoteAdapter implements Adapter {
                 // result preventing excessive calls to the Realm.
             } else {
                 // The connector isn't configured for authorization. Create a
-                // user without any roles using the supplied username.
+                // user without any roles using the supplied user name.
                 request.setUserPrincipal(new CoyotePrincipal(username));
             }
         }
@@ -909,10 +906,10 @@ public class CoyoteAdapter implements Adapter {
         // encoding that doesn't give the expected result so be explicit
         Charset charset = connector.getURICharset();
 
-        if (log.isTraceEnabled()) {
-            log.trace(sm.getString("coyoteAdapter.debug", "uriBC", uriBC.toString()));
-            log.trace(sm.getString("coyoteAdapter.debug", "semicolon", String.valueOf(semicolon)));
-            log.trace(sm.getString("coyoteAdapter.debug", "enc", charset.name()));
+        if (log.isDebugEnabled()) {
+            log.debug(sm.getString("coyoteAdapter.debug", "uriBC", uriBC.toString()));
+            log.debug(sm.getString("coyoteAdapter.debug", "semicolon", String.valueOf(semicolon)));
+            log.debug(sm.getString("coyoteAdapter.debug", "enc", charset.name()));
         }
 
         while (semicolon > -1) {
@@ -943,10 +940,10 @@ public class CoyoteAdapter implements Adapter {
                 uriBC.setEnd(start + semicolon);
             }
 
-            if (log.isTraceEnabled()) {
-                log.trace(sm.getString("coyoteAdapter.debug", "pathParamStart", String.valueOf(pathParamStart)));
-                log.trace(sm.getString("coyoteAdapter.debug", "pathParamEnd", String.valueOf(pathParamEnd)));
-                log.trace(sm.getString("coyoteAdapter.debug", "pv", pv));
+            if (log.isDebugEnabled()) {
+                log.debug(sm.getString("coyoteAdapter.debug", "pathParamStart", String.valueOf(pathParamStart)));
+                log.debug(sm.getString("coyoteAdapter.debug", "pathParamEnd", String.valueOf(pathParamEnd)));
+                log.debug(sm.getString("coyoteAdapter.debug", "pv", pv));
             }
 
             if (pv != null) {
@@ -955,10 +952,10 @@ public class CoyoteAdapter implements Adapter {
                     String name = pv.substring(0, equals);
                     String value = pv.substring(equals + 1);
                     request.addPathParameter(name, value);
-                    if (log.isTraceEnabled()) {
-                        log.trace(sm.getString("coyoteAdapter.debug", "equals", String.valueOf(equals)));
-                        log.trace(sm.getString("coyoteAdapter.debug", "name", name));
-                        log.trace(sm.getString("coyoteAdapter.debug", "value", value));
+                    if (log.isDebugEnabled()) {
+                        log.debug(sm.getString("coyoteAdapter.debug", "equals", String.valueOf(equals)));
+                        log.debug(sm.getString("coyoteAdapter.debug", "name", name));
+                        log.debug(sm.getString("coyoteAdapter.debug", "value", value));
                     }
                 }
             }
@@ -1022,8 +1019,8 @@ public class CoyoteAdapter implements Adapter {
                     request.setRequestedSessionId(scookie.getValue().toString());
                     request.setRequestedSessionCookie(true);
                     request.setRequestedSessionURL(false);
-                    if (log.isTraceEnabled()) {
-                        log.trace(" Requested cookie session id is " + request.getRequestedSessionId());
+                    if (log.isDebugEnabled()) {
+                        log.debug(" Requested cookie session id is " + request.getRequestedSessionId());
                     }
                 } else {
                     if (!request.isRequestedSessionIdValid()) {
@@ -1057,7 +1054,7 @@ public class CoyoteAdapter implements Adapter {
 
         B2CConverter conv = request.getURIConverter();
         if (conv == null) {
-            conv = new B2CConverter(charset, true);
+            conv = new B2CConverter(charset, false);
             request.setURIConverter(conv);
         } else {
             conv.recycle();
@@ -1075,7 +1072,7 @@ public class CoyoteAdapter implements Adapter {
 
 
     /**
-     * Character conversion of the US-ASCII MessageBytes.
+     * Character conversion of the a US-ASCII MessageBytes.
      *
      * @param mb The MessageBytes instance containing the bytes that should be converted to chars
      */
@@ -1106,35 +1103,39 @@ public class CoyoteAdapter implements Adapter {
     /**
      * This method normalizes "\", "//", "/./" and "/../".
      *
-     * @param uriMB URI to be normalized
+     * @param uriMB          URI to be normalized
+     * @param allowBackslash <code>true</code> if backslash characters are allowed in URLs
      *
      * @return <code>false</code> if normalizing this URI would require going above the root, or if the URI contains a
      *             null byte, otherwise <code>true</code>
      */
-    public static boolean normalize(MessageBytes uriMB) {
+    public static boolean normalize(MessageBytes uriMB, boolean allowBackslash) {
 
         ByteChunk uriBC = uriMB.getByteChunk();
         final byte[] b = uriBC.getBytes();
         final int start = uriBC.getStart();
         int end = uriBC.getEnd();
+        boolean appendedSlash = false;
 
         // An empty URL is not acceptable
         if (start == end) {
             return false;
         }
 
+        int pos = 0;
+        int index = 0;
+
+
         // The URL must start with '/' (or '\' that will be replaced soon)
         if (b[start] != (byte) '/' && b[start] != (byte) '\\') {
             return false;
         }
 
-        int pos;
-
         // Replace '\' with '/'
         // Check for null byte
         for (pos = start; pos < end; pos++) {
             if (b[pos] == (byte) '\\') {
-                if (ALLOW_BACKSLASH) {
+                if (allowBackslash) {
                     b[pos] = (byte) '/';
                 } else {
                     return false;
@@ -1161,12 +1162,13 @@ public class CoyoteAdapter implements Adapter {
             if ((b[end - 2] == (byte) '/') || ((b[end - 2] == (byte) '.') && (b[end - 3] == (byte) '/'))) {
                 b[end] = (byte) '/';
                 end++;
+                appendedSlash = true;
             }
         }
 
         uriBC.setEnd(end);
 
-        int index = 0;
+        index = 0;
 
         // Resolve occurrences of "/./" in the normalized path
         while (true) {
@@ -1203,76 +1205,16 @@ public class CoyoteAdapter implements Adapter {
             index = index2;
         }
 
-        return true;
-
-    }
-
-
-    /**
-     * Check that the URI is normalized following character decoding. This method checks for "\", 0, "//", "/./" and
-     * "/../".
-     *
-     * @param uriMB URI to be checked (should be chars)
-     *
-     * @return <code>false</code> if sequences that are supposed to be normalized are still present in the URI,
-     *             otherwise <code>true</code>
-     *
-     * @deprecated This code will be removed in Apache Tomcat 10 onwards
-     */
-    @Deprecated
-    public static boolean checkNormalize(MessageBytes uriMB) {
-
-        CharChunk uriCC = uriMB.getCharChunk();
-        char[] c = uriCC.getChars();
-        int start = uriCC.getStart();
-        int end = uriCC.getEnd();
-
-        int pos = 0;
-
-        // Check for '\' and 0
-        for (pos = start; pos < end; pos++) {
-            if (c[pos] == '\\') {
-                return false;
-            }
-            if (c[pos] == 0) {
-                return false;
-            }
-        }
-
-        // Check for "//"
-        for (pos = start; pos < (end - 1); pos++) {
-            if (c[pos] == '/') {
-                if (c[pos + 1] == '/') {
-                    return false;
-                }
-            }
-        }
-
-        // Check for ending with "/." or "/.."
-        if (((end - start) >= 2) && (c[end - 1] == '.')) {
-            if ((c[end - 2] == '/') || ((c[end - 2] == '.') && (c[end - 3] == '/'))) {
-                return false;
-            }
-        }
-
-        // Check for "/./"
-        if (uriCC.indexOf("/./", 0, 3, 0) >= 0) {
-            return false;
-        }
-
-        // Check for "/../"
-        if (uriCC.indexOf("/../", 0, 4, 0) >= 0) {
-            return false;
+        // If a slash was appended to help normalize "/." or "/.." then remove
+        // any trailing "/" from the result unless the result is "/".
+        if (appendedSlash && end > 1 && b[end - 1] == '/') {
+            uriBC.setEnd(end - 1);
         }
 
         return true;
-
     }
 
 
-    // ------------------------------------------------------ Protected Methods
-
-
     /**
      * Copy an array of bytes to a different position. Used during normalization.
      *
@@ -1284,4 +1226,87 @@ public class CoyoteAdapter implements Adapter {
     protected static void copyBytes(byte[] b, int dest, int src, int len) {
         System.arraycopy(b, src, b, dest, len);
     }
+
+
+    /*
+     * Examine URI segment by segment for 'suspicious' URIs.
+     */
+    private static boolean checkSuspiciousURIs(ByteChunk undecodedURI) {
+        byte[] bytes = undecodedURI.getBytes();
+        int start = undecodedURI.getStart();
+        int end = undecodedURI.getEnd();
+        int segmentStart = -1;
+        int segmentEnd = -1;
+
+        // Find first segment
+        segmentStart = undecodedURI.indexOf('/', 0);
+        if (segmentStart > -1) {
+            segmentEnd = undecodedURI.indexOf('/', segmentStart + 1);
+        }
+
+        while (segmentStart > -1) {
+            int pos = start + segmentStart + 1;
+
+            // Empty segment other than final segment with path parameters
+            if (segmentEnd > 0 && bytes[pos] == ';') {
+                return true;
+            }
+
+            // encoded dot-segments and/or dot-segments with path parameters
+            int dotCount = 0;
+            boolean encodedDot = false;
+            while (pos < end) {
+                if (bytes[pos] == '.') {
+                    dotCount++;
+                    pos++;
+                } else if (pos + 2 < end && bytes[pos] == '%' && bytes[pos + 1] == '2' &&
+                        (bytes[pos + 2] == 'e' || bytes[pos + 2] == 'E')) {
+                    encodedDot = true;
+                    dotCount++;
+                    pos += 3;
+                } else if (bytes[pos] == ';') {
+                    if (dotCount > 0) {
+                        return true;
+                    }
+                    break;
+                } else if (bytes[pos] == '/') {
+                    break;
+                } else {
+                    dotCount = 0;
+                    break;
+                }
+            }
+            if (dotCount > 0 && encodedDot) {
+                return true;
+            }
+
+            // %nn encoded controls or '/'
+            pos = start + segmentStart + 1;
+            while (pos < end) {
+                if (pos + 2 < end && bytes[pos] == '%') {
+                    byte b1 = bytes[pos + 1];
+                    byte b2 = bytes[pos + 2];
+                    pos += 3;
+                    int decoded = (HexUtils.getDec(b1) << 4) + HexUtils.getDec(b2);
+                    if (decoded < 20 || decoded == 0x7F || decoded == 0x2F) {
+                        return true;
+                    }
+                } else {
+                    pos++;
+                }
+            }
+
+            // Move to next segment
+            if (segmentEnd == -1) {
+                segmentStart = -1;
+            } else {
+                segmentStart = segmentEnd;
+                if (segmentStart > -1) {
+                    segmentEnd = undecodedURI.indexOf('/', segmentStart + 1);
+                }
+            }
+        }
+
+        return false;
+    }
 }
